# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import operator
from collections.abc import Iterable
from typing import TypeVar, cast

import structlog
from sqlalchemy import select
from tabulate import tabulate

import orchestrator.workflows
from orchestrator.cli.helpers.input_helpers import _enumerate_menu_keys, _prompt_user_menu, get_user_input
from orchestrator.cli.helpers.print_helpers import COLOR, noqa_print, print_fmt
from orchestrator.db import WorkflowTable, db
from orchestrator.targets import Target
from orchestrator.workflow import Workflow
from orchestrator.workflows import get_workflow

logger = structlog.get_logger(__name__)

T = TypeVar("T")


def _print_tasks_table(tasks: list[WorkflowTable]) -> None:
    items = [(task.name, task.description) for task in tasks]
    print_fmt(
        tabulate(
            items,
            headers=["#", "name", "description"],
            showindex=itertools.count(1),
        ),
        end="\n\n",
    )


def _add_task(tasks: dict[str, Workflow], state: dict) -> dict:
    print_fmt("\nAdd new task\n", flags=[COLOR.UNDERLINE])

    if not tasks:
        noqa_print("No registered tasks found to add to the database")
        return state

    noqa_print("\nWhich task should be added?")

    already_used_tasks = {task["name"] for task in state["tasks_to_add"] + state["tasks_to_delete"]}
    task_options = [(task, task) for task in tasks.keys() if task not in already_used_tasks]
    task_name = _prompt_user_menu([*task_options, ("cancel", None)], keys=[*_enumerate_menu_keys(task_options), "q"])
    if not task_name:
        # Menu cancelled
        return state

    task_to_add = {"name": task_name, "description": tasks[task_name].description}
    return {**state, "tasks_to_add": [*state["tasks_to_add"], task_to_add]}


def _delete_task(tasks: Iterable[WorkflowTable], state: dict) -> dict:
    print_fmt("\nDelete existing task\n", flags=[COLOR.UNDERLINE])
    already_used_tasks = {task["name"] for task in state["tasks_to_add"] + state["tasks_to_delete"]}
    items = [(task.name, task.description) for task in tasks if task not in already_used_tasks]
    items = sorted(items, key=operator.itemgetter(1, 1))
    keys = ["#", "name", "description"]
    if not items:
        noqa_print("No deletable tasks in database")
        return state

    print_fmt(
        tabulate(
            items,
            headers=keys,
            showindex=itertools.count(1),
        ),
        end="\n\n",
    )
    task_num = get_user_input("Which task do you want to delete? (q to cancel) ", "q")
    if not task_num.isdigit():
        return state
    task_index = int(task_num) - 1
    if 0 <= task_index < len(items):
        item = dict(zip(keys[1:], items[task_index]))
        return {**state, "tasks_to_delete": [*state["tasks_to_delete"], item]}
    return state


def _show_state(state: dict) -> dict:
    print_fmt("\nTasks to add:", flags=[COLOR.GREEN])
    print_fmt(
        tabulate(
            state["tasks_to_add"],
            headers="keys",
            showindex=itertools.count(1),
        ),
        end="\n\n",
    )

    print_fmt("Tasks to delete:", flags=[COLOR.RED])
    print_fmt(
        tabulate(
            state["tasks_to_delete"],
            headers="keys",
            showindex=itertools.count(1),
        ),
        end="\n\n",
    )
    return state


def delete_dangling_tasks(tasks: list[WorkflowTable], state: dict) -> dict:
    if not tasks:
        noqa_print("No dangling tasks found.")
        return state

    print_fmt(
        "\nThe following tasks were found in the database that do not have a corresponding LazyWorkflowInstance:\n"
    )
    _print_tasks_table(tasks)
    should_delete_dangling_tasks = (
        get_user_input("Do you wish to delete all dangling tasks from the database? [y/n]: ", "n").lower() == "y"
    )

    if not should_delete_dangling_tasks:
        noqa_print("Cancelling")
        return state

    already_used_tasks = {task["name"] for task in state["tasks_to_add"] + state["tasks_to_delete"]}
    keys = ["name", "description"]
    items = [
        {"name": task.name, "description": task.description}
        for k, task in zip(keys, tasks)
        if task.name not in already_used_tasks
    ]
    return {**state, "tasks_to_delete": [*state["tasks_to_delete"], *items]}


def create_tasks_migration_wizard() -> tuple[list[dict], list[dict]]:
    """Create tuple with lists for tasks to add and delete.

    Returns tuple:
        - list of task items to add in the migration
        - list of task items to delete in the migration
    """
    database_tasks = {task.name: task for task in list(db.session.scalars(select(WorkflowTable))) if task.is_task}
    registered_wf_instances = {
        task: cast(Workflow, get_workflow(task)) for task in orchestrator.workflows.ALL_WORKFLOWS.keys()
    }

    is_task = [Target.SYSTEM, Target.VALIDATE]

    registered_tasks = dict(
        filter(
            lambda task: task[1].target in is_task and task[0] in database_tasks.keys(),
            registered_wf_instances.items(),
        )
    )
    available_tasks = dict(
        filter(
            lambda task: task[1].target in is_task and task[0] not in database_tasks.keys(),
            registered_wf_instances.items(),
        )
    )
    unregistered_tasks = [task for task in database_tasks.values() if task.name not in registered_tasks.keys()]

    # Main menu loop
    state = {"tasks_to_add": [], "tasks_to_delete": [], "done": False}
    while not state["done"]:
        print_fmt("\nWhat do you want to do?\n", flags=[COLOR.UNDERLINE, COLOR.BOLD])
        choice_fn = _prompt_user_menu(
            [
                ("Add task to database", lambda s: _add_task(available_tasks, s)),
                ("Delete task from database", lambda s: _delete_task(database_tasks.values(), s)),
                (
                    "Delete unregistered tasks from database",
                    lambda s: delete_dangling_tasks(unregistered_tasks, s),
                ),
                ("Finish and create migration file", lambda s: {**s, "done": True, "abort": False}),
                ("Show current diff", _show_state),
                ("Quit menu without creating a migration file", lambda s: {**s, "done": True, "abort": True}),
            ],
            keys=["a", "x", "c", "y", "d", "q"],
        )
        if choice_fn:
            state = choice_fn(state)  # type: ignore

    if state.get("abort"):
        return [], []

    logger.info("Create tasks", create_tasks=state["tasks_to_add"])
    logger.info("Delete tasks", delete_tasks=state["tasks_to_delete"])

    return state["tasks_to_add"], state["tasks_to_delete"]  # type: ignore



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import HTTPStatus

import structlog
from fastapi.routing import APIRouter
from sqlalchemy import select
from sqlalchemy.exc import OperationalError

from orchestrator.api.error_handling import raise_status
from orchestrator.db import ProductTable, db

logger = structlog.get_logger(__name__)

router = APIRouter()


@router.get("/", response_model=str)
def get_health() -> str:
    try:
        stmt = select(ProductTable.name).limit(1)
        db.session.execute(stmt)
    except OperationalError as e:
        logger.warning("Health endpoint returned: notok!")
        logger.debug("Health endpoint error details", error=str(e))
        raise_status(HTTPStatus.INTERNAL_SERVER_ERROR)
    return "OK"



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial
from uuid import UUID

import structlog
from celery import Celery, Task
from celery.app.control import Inspect
from celery.utils.log import get_task_logger
from kombu.serialization import registry

from orchestrator.schemas.engine_settings import WorkerStatus
from orchestrator.services.executors.threadpool import thread_resume_process, thread_start_process
from orchestrator.services.processes import _get_process, ensure_correct_process_status, load_process
from orchestrator.types import BroadcastFunc
from orchestrator.utils.json import json_dumps, json_loads
from orchestrator.workflow import ProcessStatus

logger = get_task_logger(__name__)

local_logger = structlog.get_logger(__name__)

_celery: Celery | None = None

NEW_TASK = "tasks.new_task"
NEW_WORKFLOW = "tasks.new_workflow"
RESUME_TASK = "tasks.resume_task"
RESUME_WORKFLOW = "tasks.resume_workflow"


def get_celery_task(task_name: str) -> Task:
    if _celery:
        return _celery.signature(task_name)
    raise AssertionError("Celery has not been initialised yet")


def register_custom_serializer() -> None:
    # orchestrator specific serializer to correctly handle more complex classes
    registry.register("orchestrator-json", json_dumps, json_loads, "application/json", "utf-8")


def initialise_celery(celery: Celery) -> None:  # noqa: C901
    global _celery
    if _celery:
        raise AssertionError("You can only initialise Celery once")
    _celery = celery

    # Different routes/queues so we can assign them priorities
    celery.conf.task_routes = {
        NEW_TASK: {"queue": "new_tasks"},
        NEW_WORKFLOW: {"queue": "new_workflows"},
        RESUME_TASK: {"queue": "resume_tasks"},
        RESUME_WORKFLOW: {"queue": "resume_workflows"},
    }

    register_custom_serializer()

    process_broadcast_fn: BroadcastFunc | None = getattr(celery, "process_broadcast_fn", None)

    def start_process(process_id: UUID, user: str) -> UUID | None:
        try:
            process = _get_process(process_id)
            pstat = load_process(process)
            ensure_correct_process_status(process_id, ProcessStatus.CREATED)
            thread_start_process(pstat, user)

        except Exception as exc:
            local_logger.error("Worker failed to execute workflow", process_id=process_id, details=str(exc))
            return None
        else:
            return process_id

    def resume_process(process_id: UUID, user: str) -> UUID | None:
        try:
            process = _get_process(process_id)
            ensure_correct_process_status(process_id, ProcessStatus.RESUMED)
            thread_resume_process(process, user=user, broadcast_func=process_broadcast_fn)
        except Exception as exc:
            local_logger.error("Worker failed to resume workflow", process_id=process_id, details=str(exc))
            return None
        else:
            return process_id

    celery_task = partial(celery.task, log=local_logger, serializer="orchestrator-json")

    @celery_task(name=NEW_TASK)  # type: ignore
    def new_task(process_id: UUID, user: str) -> UUID | None:
        local_logger.info("Start task", process_id=process_id)
        return start_process(process_id, user=user)

    @celery_task(name=NEW_WORKFLOW)  # type: ignore
    def new_workflow(process_id: UUID, user: str) -> UUID | None:
        local_logger.info("Start workflow", process_id=process_id)
        return start_process(process_id, user=user)

    @celery_task(name=RESUME_TASK)  # type: ignore
    def resume_task(process_id: UUID, user: str) -> UUID | None:
        local_logger.info("Resume task", process_id=process_id)
        return resume_process(process_id, user=user)

    @celery_task(name=RESUME_WORKFLOW)  # type: ignore
    def resume_workflow(process_id: UUID, user: str) -> UUID | None:
        local_logger.info("Resume workflow", process_id=process_id)
        return resume_process(process_id, user=user)


class CeleryJobWorkerStatus(WorkerStatus):
    def __init__(self) -> None:
        super().__init__(executor_type="celery")
        if not _celery:
            logger.error("Can't create CeleryJobStatistics. Celery is not initialised.")
            return

        inspection: Inspect = _celery.control.inspect()
        stats = inspection.stats()
        self.number_of_workers_online = len(stats)

        def sum_items(d: dict) -> int:
            return sum(len(lines) for _, lines in d.items()) if d else 0

        self.number_of_queued_jobs = sum_items(inspection.scheduled()) + sum_items(inspection.reserved())
        self.number_of_running_jobs = sum(len(tasks) for w, tasks in inspection.active().items())



from typing import Literal
from uuid import UUID

import structlog
from more_itertools import one, unique_everseen
from strawberry.dataloader import DataLoader

from orchestrator.db import (
    SubscriptionTable,
)
from orchestrator.services.subscription_relations import get_depends_on_subscriptions, get_in_use_by_subscriptions
from orchestrator.types import SubscriptionLifecycle

logger = structlog.get_logger(__name__)


def unzip_subscription_ids_and_filter_statuses(
    keys: list[tuple[UUID, tuple[str, ...]]], direction: Literal["dependsOn", "inUseBy"]
) -> tuple[list[UUID], tuple[str, ...]]:
    subscription_ids = [key[0] for key in keys]
    filter_statuses_values = (key[1] for key in keys)
    filter_statuses = one(
        unique_everseen(filter_statuses_values),
        too_long=Exception(f"{direction}Filter.statuses must be set to the same value"),
    )
    return subscription_ids, tuple(filter_statuses or SubscriptionLifecycle.values())


async def in_use_by_subs_loader(keys: list[tuple[UUID, tuple[str, ...]]]) -> list[list[SubscriptionTable]]:
    """GraphQL dataloader to efficiently get the in_use_by SubscriptionTables for multiple subscription_ids."""
    subscription_ids, filter_statuses = unzip_subscription_ids_and_filter_statuses(keys, "inUseBy")
    return await get_in_use_by_subscriptions(subscription_ids, filter_statuses)


async def depends_on_subs_loader(keys: list[tuple[UUID, tuple[str, ...]]]) -> list[list[SubscriptionTable]]:
    """GraphQL dataloader to efficiently get the depends_on SubscriptionTables for multiple subscription_ids."""
    subscription_ids, filter_statuses = unzip_subscription_ids_and_filter_statuses(keys, "inUseBy")
    return await get_depends_on_subscriptions(subscription_ids, filter_statuses)


SubsLoaderType = DataLoader[tuple[UUID, tuple[str, ...]], list[SubscriptionTable]]



# Copyright 2019-2025 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Literal
from uuid import UUID

import structlog
from sqlalchemy import select

from orchestrator.db import db
from orchestrator.db.models import InputStateTable

logger = structlog.get_logger(__name__)

InputType = Literal["initial_state", "user_input"]


def retrieve_input_state(process_id: UUID, input_type: InputType, raise_exception: bool = True) -> InputStateTable:
    """Get user input.

    Args:
        process_id: Process ID
        input_type: The type of the input.
        raise_exception: boolean to throw error when not finding data or not

    Returns:
        User input table

    """

    res: InputStateTable | None = db.session.scalars(
        select(InputStateTable)
        .filter(InputStateTable.process_id == process_id)
        .filter(InputStateTable.input_type == input_type)
        .order_by(InputStateTable.input_time.asc())
    ).first()

    if res:
        logger.debug("Retrieved input state", process_id=process_id, input_state=res, input_type=input_type)
        return res
    if raise_exception:
        raise ValueError(f"No input state for pid: {process_id}")
    return InputStateTable(input_state={})


def store_input_state(
    process_id: UUID,
    input_state: dict[str, Any] | list[dict[str, Any]],
    input_type: InputType,
) -> None:
    """Store user input state.

    Args:
        process_id: Process ID
        input_state: Dictionary of user input state
        input_type: The type of the input.

    Returns:
        None

    """
    logger.debug("Store input state", process_id=process_id, input_state=input_state, input_type=input_type)
    db.session.add(
        InputStateTable(
            process_id=process_id,
            input_state=input_state,
            input_type=input_type,
        )
    )
    db.session.commit()



"""Fix tsv triggers.

Revision ID: 3323bcb934e7
Revises: c112305b07d3
Create Date: 2020-12-10 09:22:46.491454

"""

from alembic import op
from sqlalchemy import text

# revision identifiers, used by Alembic.
revision = "3323bcb934e7"
down_revision = "a76b9185b334"
branch_labels = None
depends_on = None


def upgrade() -> None:
    conn = op.get_bind()
    conn.execute(
        text(
            """
            DROP TRIGGER IF EXISTS subscriptions_upd_trigger ON subscriptions;

            CREATE OR REPLACE FUNCTION subscriptions_upd_trigger() RETURNS trigger
                LANGUAGE plpgsql
                AS $$
            BEGIN
                UPDATE subscriptions
                SET tsv = NULL
                WHERE subscription_id = NEW.subscription_id;
                RETURN NULL;
            END
            $$;

            CREATE TRIGGER subscriptions_upd_trigger
            AFTER UPDATE
            ON public.subscriptions
            FOR EACH ROW
                /*

                After updating a row on `subscriptions` we will perform another update on that same row to set the value of the
                `tsv` column. That has the potential to result in a cascade of triggers unless we specifically guard against it.

                This trigger will fire when anything but the tsv column changes and will set tsv to NULL.

                */
            WHEN (old.tsv IS NOT DISTINCT FROM new.tsv)
            EXECUTE FUNCTION public.subscriptions_upd_trigger();

            CREATE OR REPLACE FUNCTION subscriptions_set_tsv_trigger() RETURNS trigger
                LANGUAGE plpgsql
                AS $$
            BEGIN
                UPDATE subscriptions
                SET tsv = generate_subscription_tsv(NEW.subscription_id)
                WHERE subscription_id = NEW.subscription_id;
                RETURN NULL;
            END
            $$;

            CREATE TRIGGER subscriptions_set_tsv_trigger
            AFTER UPDATE
            ON public.subscriptions
            FOR EACH ROW
                /*

                After updating a row on `subscriptions` we will perform another update on that same row to set the value of the
                `tsv` column. That has the potential to result in a cascade of triggers unless we specifically guard against it.

                This trigger will only fire when tsv is set to NULL. This happens when something in another table or in this tables changes.
                Except for changes to the tsv column itself

                */
            WHEN (new.tsv IS NULL)
            EXECUTE FUNCTION public.subscriptions_set_tsv_trigger();
        """
        )
    )


def downgrade() -> None:
    pass



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from uuid import UUID

from pydantic import ConfigDict

from orchestrator.schemas.base import OrchestratorBaseModel

TagConfig = dict[str, list[dict[str, bool]]]


class FixedInputBaseSchema(OrchestratorBaseModel):
    fixed_input_id: UUID | None = None
    name: str
    value: str
    product_id: UUID | None = None


class FixedInputSchema(FixedInputBaseSchema):
    fixed_input_id: UUID
    created_at: datetime
    product_id: UUID
    model_config = ConfigDict(from_attributes=True)


class FixedInputConfigurationItemSchema(OrchestratorBaseModel):
    name: str
    description: str
    values: list[str]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from pathlib import Path
from typing import Union

from orchestrator.settings import app_settings

Translations = dict[str, Union[dict, str]]


def _load_translations_file(language: str, translations_dir: Path) -> Translations:
    filename = translations_dir / f"{language}.json"
    if not filename.exists():
        return {}

    with filename.open() as translation_file:
        data = json.load(translation_file)
        if not isinstance(data, dict):
            return {}
        return data


def _deep_merge_dict(d1: dict, d2: dict) -> Translations:
    for k, v in d2.items():
        if isinstance(d1.get(k), dict) and isinstance(v, dict):
            d1[k] = _deep_merge_dict(d1[k], v)
        else:
            d1[k] = v
    return d1


def generate_translations(language: str) -> Translations:
    translations = _load_translations_file(language, Path(__file__).parent.parent / "workflows" / "translations")
    user_translations = (
        _load_translations_file(language, app_settings.TRANSLATIONS_DIR) if app_settings.TRANSLATIONS_DIR else {}
    )
    return _deep_merge_dict(translations, user_translations)



# Copyright 2019-2025 SURF, GÉANT, ESnet.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Callable, Sequence
from concurrent.futures.thread import ThreadPoolExecutor
from datetime import datetime
from functools import partial
from http import HTTPStatus
from typing import Any
from uuid import UUID, uuid4

import structlog
from deepmerge.merger import Merger
from pytz import utc
from sqlalchemy import delete, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import joinedload

from nwastdlib.ex import show_ex
from oauth2_lib.fastapi import OIDCUserModel
from orchestrator.api.error_handling import raise_status
from orchestrator.config.assignee import Assignee
from orchestrator.db import EngineSettingsTable, ProcessStepTable, ProcessSubscriptionTable, ProcessTable, db
from orchestrator.distlock import distlock_manager
from orchestrator.schemas.engine_settings import WorkerStatus
from orchestrator.services.input_state import store_input_state
from orchestrator.services.settings import get_engine_settings_for_update
from orchestrator.services.workflows import get_workflow_by_name
from orchestrator.settings import ExecutorType, app_settings
from orchestrator.types import BroadcastFunc
from orchestrator.utils.datetime import nowtz
from orchestrator.utils.errors import error_state_to_dict
from orchestrator.websocket import broadcast_invalidate_status_counts, broadcast_process_update_to_websocket
from orchestrator.workflow import (
    CALLBACK_TOKEN_KEY,
    DEFAULT_CALLBACK_PROGRESS_KEY,
    Failed,
    ProcessStat,
    ProcessStatus,
    Step,
    StepList,
    Success,
    Workflow,
    abort_wf,
)
from orchestrator.workflow import Process as WFProcess
from orchestrator.workflows import get_workflow
from orchestrator.workflows.removed_workflow import removed_workflow
from pydantic_forms.core import post_form
from pydantic_forms.exceptions import FormValidationError
from pydantic_forms.types import State

logger = structlog.get_logger(__name__)

StateMerger = Merger([(dict, ["merge"])], ["override"], ["override"])

SYSTEM_USER = "SYSTEM"

_workflow_executor = None

START_WORKFLOW_REMOVED_ERROR_MSG = "This workflow cannot be started because it has been removed"
RESUME_WORKFLOW_REMOVED_ERROR_MSG = "This workflow cannot be resumed because it has been removed"


def get_execution_context() -> dict[str, Callable]:
    if app_settings.EXECUTOR == ExecutorType.WORKER:
        from orchestrator.services.executors.celery import CELERY_EXECUTION_CONTEXT

        return CELERY_EXECUTION_CONTEXT

    from orchestrator.services.executors.threadpool import THREADPOOL_EXECUTION_CONTEXT

    return THREADPOOL_EXECUTION_CONTEXT


def get_thread_pool() -> ThreadPoolExecutor:
    """Get and optionally initialise a ThreadPoolExecutor.

    Returns:
        ThreadPoolExecutor

    """
    global _workflow_executor
    if _workflow_executor is None:
        _workflow_executor = ThreadPoolExecutor(max_workers=app_settings.MAX_WORKERS)
    return _workflow_executor


def shutdown_thread_pool() -> None:
    """Gracefully shutdown existing ThreadPoolExecutor and delete it."""
    global _workflow_executor
    if isinstance(_workflow_executor, ThreadPoolExecutor):
        _workflow_executor.shutdown(wait=True)
        _workflow_executor = None


def _db_create_process(stat: ProcessStat) -> None:
    wf_table = get_workflow_by_name(stat.workflow.name)
    if not wf_table:
        raise AssertionError(f"No workflow found with name: {stat.workflow.name}")

    p = ProcessTable(
        process_id=stat.process_id,
        workflow_id=wf_table.workflow_id,
        last_status=ProcessStatus.CREATED,
        created_by=stat.current_user,
        is_task=wf_table.is_task,
    )
    db.session.add(p)
    db.session.commit()


def delete_process(process_id: UUID) -> None:
    db.session.execute(delete(ProcessTable).where(ProcessTable.process_id == process_id))
    broadcast_invalidate_status_counts()
    db.session.commit()


def _update_process(process_id: UUID, step: Step, process_state: WFProcess) -> ProcessTable:
    p = db.session.get(ProcessTable, process_id)
    if p is None:
        raise ValueError(f"Failed to write failure step to process: process with PID {process_id} not found")

    p.last_step = step.name
    p.last_status = process_state.overall_status
    p.assignee = step.assignee
    step_state: State = process_state.unwrap()
    if process_state.isfailed() or process_state.iswaiting():
        failed_reason = step_state.get("error")
        # pop also removes the traceback from the dict
        traceback = step_state.pop("traceback", None)

        p.failed_reason = failed_reason
        p.traceback = traceback

        if process_state.isfailed() and p.is_task:
            # Check if we need a special failed status:
            # If it is an AssertionError:
            if step_state.get("class") == "AssertionError" or step_state.get("class") == "InconsistentData":
                p.assignee = Assignee.NOC
                p.last_status = ProcessStatus.INCONSISTENT_DATA
            # If we encounter a connectivity issue with an underlying api:
            elif step_state.get("class") == "MaxRetryError" or (
                step_state.get("class") == "ApiException"
                and step_state.get("status_code")
                in (
                    HTTPStatus.BAD_GATEWAY,
                    HTTPStatus.SERVICE_UNAVAILABLE,
                    HTTPStatus.GATEWAY_TIMEOUT,
                )
            ):
                p.assignee = Assignee.SYSTEM
                p.last_status = ProcessStatus.API_UNAVAILABLE
            else:
                p.assignee = Assignee.SYSTEM

    else:
        p.failed_reason = None
        p.traceback = None

    return p


def ensure_correct_process_status(process_id: UUID, last_status: str) -> None:
    p = db.session.get(ProcessTable, process_id)
    if p is None:
        raise ValueError(f"Process with id {process_id} does not exist")
    if p.last_status != last_status:
        raise AssertionError(f"Process {process_id} has last_status {p.last_status}, expected: {last_status}")


def _get_current_step_to_update(
    stat: ProcessStat, p: ProcessTable, step: Step, process_state: WFProcess
) -> ProcessStepTable:
    """Checks if last error state is identical to determine if we add a new step or update the last one.

    There are some special internal keys checked to customise how the step is updated in the database.
    These keys are only used for this purpose and not stored in the database.
    - '__step_name_override' - When this key is present, the value will be used as the ProcessStepTable.name
    - '__replace_last_state' - When the value for this key is truthy, the previous state in the db will be overridden
    - '__remove_keys' - Removes the keys in this given list from the state
    """
    step_state: State = process_state.unwrap()
    current_step = None
    last_db_step = p.steps[-1] if len(p.steps) else None

    # Core internal: __step_name_override
    step_name = step_state.pop("__step_name_override", step.name)

    # Core internal: __replace_last_state
    if step_state.pop("__replace_last_state", None):
        current_step = last_db_step
        current_step.status = process_state.status
        current_step.state = step_state

    # Core internal: __remove_keys
    try:
        keys_to_remove = step_state.get("__remove_keys", [])
        for k in keys_to_remove:
            step_state.pop(k, None)
    except TypeError:
        logger.error("Value for '__keys_to_remove' is not iterable.")
    finally:
        step_state.pop("__remove_keys", None)

    # We don't have __last_step_started in __remove_keys because the way __remove_keys is populated appears like it would overwrite
    # what's put there in the step decorator in certain cases (step groups and callback steps)
    step_start_time = step_state.pop("__last_step_started_at", None)

    if process_state.isfailed() or process_state.iswaiting():
        if (
            last_db_step is not None
            and last_db_step.status == process_state.status
            and last_db_step.name == step.name
            and last_db_step.state.get("error") == step_state.get("error")
            and last_db_step.state.get("details") == step_state.get("details")
        ):
            state_ex_info = {
                "retries": last_db_step.state.get("retries", 0) + 1,
                "completed_at": last_db_step.state.get("completed_at", []) + [str(last_db_step.completed_at)],
            }

            # write new state info and execution date
            last_db_step.state = step_state | state_ex_info
            logger.info(
                "Updating existing process step with state info about the error",
                retries=state_ex_info["retries"],
            )
            current_step = last_db_step

    if current_step is None:
        # add a new entry to the process stat
        current_step = ProcessStepTable(
            process_id=stat.process_id,
            name=step_name,
            status=process_state.status,
            state=step_state,
            created_by=stat.current_user,
        )
    # Since the Start step does not have a __last_step_started_at in it's state, we effectively assume it is instantaneous.
    now = nowtz()
    current_step.started_at = datetime.fromtimestamp(step_start_time or now.timestamp(), tz=utc)

    # Always explicitly set this instead of leaving it to the database to prevent failing tests
    # Test will fail if multiple steps have the same timestamp
    current_step.completed_at = now
    return current_step


def _db_log_step(
    stat: ProcessStat,
    step: Step,
    process_state: WFProcess,
    broadcast_func: BroadcastFunc | None = None,
) -> WFProcess:
    """Write the current step to the db.

    Args:
        stat: ProcessStat of process
        step: Current step
        process_state: State of process after current step
        broadcast_func: Optional function to broadcast process data

    Returns:
        WFProcess: The process as stored in the database

    """
    p = _update_process(stat.process_id, step, process_state)
    current_step = _get_current_step_to_update(stat, p, step, process_state)

    db.session.add(p)
    db.session.add(current_step)
    try:
        db.session.commit()
    except BaseException:
        db.session.rollback()
        raise

    if broadcast_func:
        broadcast_func(p.process_id)

    # Return the state as stored in the database
    return process_state.__class__(current_step.state)


def safe_logstep(
    stat: ProcessStat,
    step: Step,
    process_state: WFProcess,
    broadcast_func: BroadcastFunc | None = None,
) -> WFProcess:
    """Log step and handle failures in logging.

    We need to be robust in failures to write steps to database. If that happens we try again with the failure.
    If that is also failing we give up by raising an exception which should be caught and written by _db_log_process_ex

    If the broadcast_func raises an exception, the process will go to a failed state.
    """

    try:
        return _db_log_step(stat, step, process_state, broadcast_func=broadcast_func)
    except Exception as e:
        logger.exception("Failed to save step", stat=stat, step=step, process_state=process_state)
        failure = Failed(error_state_to_dict(e))

        # Try writing the failure, but return the original exception on success
        # on a second failure the exception should be handled higher
        return _db_log_step(stat, step, failure, broadcast_func=broadcast_func)


def _db_log_process_ex(process_id: UUID, ex: Exception) -> None:
    """Write the exception to the process or task when everything else has failed.

    Args:
        process_id: the process_id of the workflow process
        ex: the Exception message

    Returns: None, there is no one to listen at this point

    """

    p = db.session.get(ProcessTable, process_id)
    if p is None:
        logger.error(
            "Failed to write failure to database: Process with PID %s not found",
            process_id,
            process_id=process_id,
        )
        return

    logger.warning("Writing only process state to DB as step couldn't be found", process_id=process_id)
    p.last_step = "Unknown"
    if p.last_status != ProcessStatus.WAITING:
        p.last_status = ProcessStatus.FAILED
    p.failed_reason = str(ex)
    p.traceback = show_ex(ex)
    db.session.add(p)
    try:
        db.session.commit()
    except BaseException:
        logger.exception("Commit failed, rolling back", process_id=process_id)
        db.session.rollback()
        raise


def _get_process(process_id: UUID) -> ProcessTable:
    process = db.session.get(
        ProcessTable,
        process_id,
        options=[
            joinedload(ProcessTable.steps),
            joinedload(ProcessTable.process_subscriptions).joinedload(ProcessSubscriptionTable.subscription),
        ],
    )

    if not process:
        raise_status(HTTPStatus.NOT_FOUND, f"Process with process_id {process_id} not found")

    return process


def _run_process_async(process_id: UUID, f: Callable) -> UUID:
    def _update_running_processes(method: str, *args: Any) -> None:
        """Update amount of running processes by one.

        Args:
            method: Add or subtract by one the amount of running processes
            args: Any args that are still going to be passed. When called as a callback this will be the future.

        Returns:
            None

        """
        engine_settings = get_engine_settings_for_update()
        engine_settings.running_processes += 1 if method == "+" else -1
        if engine_settings.running_processes < 0:
            engine_settings.running_processes = 0
        db.session.commit()

    def run() -> WFProcess:
        try:
            with db.database_scope():
                try:
                    logger.new(process_id=str(process_id))
                    result = f()
                except Exception as ex:
                    # We still have access to the database, so we can log at least something
                    _db_log_process_ex(process_id, ex)
                    raise
                finally:
                    _update_running_processes("-")
        except Exception as ex:
            # We lost access to database here, so we can only log
            logger.exception("Unknown workflow failure", process_id=process_id)
            result = Failed(ex)

        return result

    _update_running_processes("+")
    if app_settings.EXECUTOR == ExecutorType.THREADPOOL:
        workflow_executor = get_thread_pool()
        process_handle = workflow_executor.submit(run)

        # Wait for the thread to return.
        if app_settings.TESTING:
            process_handle.result()
    elif app_settings.EXECUTOR == ExecutorType.WORKER:
        # No need to run in a thread, just run.
        run()
    else:
        raise RuntimeError("Unknown Executor type")
    return process_id


def error_message_unauthorized(workflow_key: str) -> str:
    return f"User is not authorized to execute '{workflow_key}' workflow"


def create_process(
    workflow_key: str,
    user_inputs: list[State] | None = None,
    user: str = SYSTEM_USER,
    user_model: OIDCUserModel | None = None,
) -> ProcessStat:
    # ATTENTION!! When modifying this function make sure you make similar changes to `run_workflow` in the test code

    if user_inputs is None:
        user_inputs = [{}]

    process_id = uuid4()
    workflow = get_workflow(workflow_key)

    if not workflow:
        raise_status(HTTPStatus.NOT_FOUND, "Workflow does not exist")

    if not workflow.authorize_callback(user_model):
        raise_status(HTTPStatus.FORBIDDEN, error_message_unauthorized(workflow_key))

    initial_state = {
        "process_id": process_id,
        "reporter": user,
        "workflow_name": workflow_key,
        "workflow_target": workflow.target,
    }

    try:
        state = post_form(workflow.initial_input_form, initial_state, user_inputs)
    except FormValidationError:
        logger.exception("Validation errors", user_inputs=user_inputs)
        raise

    pstat = ProcessStat(
        process_id,
        workflow=workflow,
        state=Success(state | initial_state),
        log=workflow.steps,
        current_user=user,
        user_model=user_model,
    )

    _db_create_process(pstat)
    store_input_state(process_id, state | initial_state, "initial_state")
    return pstat


def start_process(
    workflow_key: str,
    user_inputs: list[State] | None = None,
    user: str = SYSTEM_USER,
    user_model: OIDCUserModel | None = None,
    broadcast_func: BroadcastFunc | None = None,
) -> UUID:
    """Start a process for workflow.

    Args:
        workflow_key: name of workflow
        user_inputs: List of form inputs from frontend
        user: User who starts this process
        user_model: Full OIDCUserModel with claims, etc
        broadcast_func: Optional function to broadcast process data

    Returns:
        process id

    """
    pstat = create_process(workflow_key, user_inputs=user_inputs, user=user)

    start_func = get_execution_context()["start"]
    return start_func(pstat, user=user, user_model=user_model, broadcast_func=broadcast_func)


def restart_process(
    process: ProcessTable,
    *,
    user: str | None = None,
    broadcast_func: BroadcastFunc | None = None,
) -> UUID:
    """Restart a process that is stuck on status CREATED.

    Args:
        process: Process from database
        user: user who resumed this process
        broadcast_func: Optional function to broadcast process data

    Returns:
        process id

    """
    pstat = load_process(process)

    start_func = get_execution_context()["start"]
    return start_func(pstat, user=user, broadcast_func=broadcast_func)


RESUMABLE_STATUSES = (
    ProcessStatus.SUSPENDED,  # Can be resumed
    ProcessStatus.WAITING,  # Can be retried
    ProcessStatus.FAILED,  # Can be retried
    ProcessStatus.API_UNAVAILABLE,  # subtype of FAILED
    ProcessStatus.INCONSISTENT_DATA,  # subtype of FAILED
    ProcessStatus.RESUMED,  # re-resume stuck process
)


def can_be_resumed(status: ProcessStatus) -> bool:
    return status in RESUMABLE_STATUSES


def resume_process(
    process: ProcessTable,
    *,
    user_inputs: list[State] | None = None,
    user: str | None = None,
    broadcast_func: BroadcastFunc | None = None,
) -> UUID:
    """Resume a failed or suspended process.

    Args:
        process: Process from database
        user_inputs: Optional user input from forms
        user: user who resumed this process
        broadcast_func: Optional function to broadcast process data

    Returns:
        process id

    """
    pstat = load_process(process)

    if pstat.workflow == removed_workflow:
        raise ValueError(RESUME_WORKFLOW_REMOVED_ERROR_MSG)

    try:
        user_input = post_form(pstat.log[0].form, pstat.state.unwrap(), user_inputs=user_inputs or [{}])
    except FormValidationError:
        logger.exception("Validation errors", user_inputs=user_inputs)
        raise

    store_input_state(pstat.process_id, user_input, "user_input")

    resume_func = get_execution_context()["resume"]
    return resume_func(process, user=user, broadcast_func=broadcast_func)


def ensure_correct_callback_token(pstat: ProcessStat, *, token: str) -> None:
    """Ensure that a callback token matches the expected value in state.

    Args:
        pstat: ProcessStat of process.
        token: The token which was generated for the process.

    Raises:
        AssertionError: if the supplied token does not match the generated process token.

    """
    state = pstat.state.unwrap()

    # Check if the token matches
    token_from_state = state.get(CALLBACK_TOKEN_KEY)
    if token != token_from_state:
        raise AssertionError("Invalid token")


def replace_current_step_state(process: ProcessTable, *, new_state: State) -> None:
    """Replace the state of the current step in a process.

    Args:
        process: Process from database
        new_state: The new state

    """
    current_step = process.steps[-1]
    current_step.state = new_state
    db.session.add(current_step)
    db.session.commit()


def continue_awaiting_process(
    process: ProcessTable,
    *,
    token: str,
    input_data: State,
    broadcast_func: BroadcastFunc | None = None,
) -> UUID:
    """Continue a process awaiting data from a callback.

    Args:
        process: Process from database
        token: The token which was generated for the process. This must match.
        input_data: Data posted to the callback
        broadcast_func: Optional function to broadcast process data

    Returns:
        process id

    """

    pstat = load_process(process)
    state = pstat.state.unwrap()

    ensure_correct_callback_token(pstat, token=token)

    # We need to pass the callback data to the worker executor. Currently, this is not supported.
    # Therefore, we update the step state in the db and kick-off resume_workflow
    # Possible improvement: Allow passing additional data to be merged to the state upon resume_workflow
    result_key = state.get("__callback_result_key", "callback_result")
    state = {**state, result_key: input_data}

    replace_current_step_state(process, new_state=state)

    # Continue the workflow
    resume_func = get_execution_context()["resume"]
    return resume_func(process, broadcast_func=broadcast_func)


def update_awaiting_process_progress(
    process: ProcessTable,
    *,
    token: str,
    data: str | State,
) -> UUID:
    """Update progress for a process awaiting data from a callback.

    Args:
        process: Process from database
        token: The token which was generated for the process. This must match.
        data: Progress data posted to the callback

    Returns:
        process id

    Raises:
        AssertionError: if the supplied token does not match the generated process token.

    """
    pstat = load_process(process)

    ensure_correct_callback_token(pstat, token=token)

    state = pstat.state.unwrap()
    progress_key = DEFAULT_CALLBACK_PROGRESS_KEY
    state = {**state, progress_key: data} | {"__remove_keys": [progress_key]}

    replace_current_step_state(process, new_state=state)
    broadcast_process_update_to_websocket(process.process_id)

    return process.process_id


async def _async_resume_processes(
    processes: Sequence[ProcessTable],
    user_name: str,
    broadcast_func: Callable | None = None,
) -> bool:
    """Asynchronously resume multiple failed processes.

    Args:
        processes: Processes from database
        user_name: User who requested resuming the processes
        broadcast_func: The broadcast functionality

    Returns:
        True if the resume-all operation has been started.
        False if it has not been started because it is already running.

    """
    lock_expiration = max(30, len(processes) // 10)
    if not (lock := await distlock_manager.get_lock("resume-all", lock_expiration)):
        return False

    def run() -> None:
        try:
            for _proc in processes:
                try:
                    process = _get_process(_proc.process_id)
                    if process.last_status == ProcessStatus.RUNNING:
                        # Process has been started by something else in the meantime
                        logger.info("Cannot resume a running process", process_id=_proc.process_id)
                        continue
                    elif process.last_status == ProcessStatus.RESUMED:  # noqa: RET507
                        # Process has been resumed by something else in the meantime
                        logger.info("Cannot resume a resumed process", process_id=_proc.process_id)
                        continue
                    resume_process(process, user=user_name, broadcast_func=broadcast_func)
                except Exception:
                    logger.exception("Failed to resume process", process_id=_proc.process_id)
            logger.info("Completed resuming processes")
        finally:
            distlock_manager.release_sync(lock)

    # Start all jobs in the background. BackgroundTasks might be more suited.
    workflow_executor = get_thread_pool()
    process_handle = workflow_executor.submit(run)
    if app_settings.TESTING:
        process_handle.result()
    return True


def abort_process(process: ProcessTable, user: str, broadcast_func: Callable | None = None) -> WFProcess:
    pstat = load_process(process)

    pstat.update(current_user=user)
    return abort_wf(pstat, partial(safe_logstep, broadcast_func=broadcast_func))


def _recoverwf(wf: Workflow, log: list[WFProcess]) -> tuple[WFProcess, StepList]:
    # Remove all extra steps (Failed, Suspended and (A)waiting steps in db). Only keep cleared steps.

    persistent = list(
        filter(lambda p: not (p.isfailed() or p.issuspend() or p.iswaiting() or p.isawaitingcallback()), log)
    )
    stepcount = len(persistent)

    if log and (log[-1].issuspend() or log[-1].isawaitingcallback()):
        # Use the state from the suspended/awaiting steps
        state = log[-1]
    elif persistent:
        # Otherwise, use the state from the last cleared step.
        state = persistent[-1]
    else:
        state = Success({})

    # Calculate the remaining Step functions that still need to be executed.
    # If the StepList has changed since the process was started, it is not possible to do this perfectly based on the ProcessSteps log in the database.
    # It is not advisable to add or remove a step from a workflow other than at the end, or a running workflow may skip or re-execute a step and will probably result in errors.
    # Adding or removing a step at the end (before the `done` step) is ok, because these steps have not run yet.
    # A step with status Completed will always be treated as the last step.
    rest = StepList() if state.iscomplete() else wf.steps[stepcount:]

    return state, rest


def _restore_log(steps: list[ProcessStepTable]) -> list[WFProcess]:
    result = []
    for step in steps:
        process = WFProcess.from_status(step.status, step.state)

        if not process:
            raise ValueError(step.status)

        result.append(process)
    return result


def load_process(process: ProcessTable) -> ProcessStat:
    workflow = get_workflow(str(process.workflow.name))

    if not workflow:
        workflow = removed_workflow

    log = _restore_log(process.steps)
    pstate, remaining = _recoverwf(workflow, log)

    return ProcessStat(
        process_id=process.process_id,
        workflow=workflow,
        state=pstate,
        log=remaining,
        current_user=SYSTEM_USER,
    )


def _get_running_processes() -> list[ProcessTable]:
    stmt = select(ProcessTable).where(ProcessTable.last_status == "running")
    return list(db.session.scalars(stmt))


def set_process_status(process: ProcessTable, status: ProcessStatus) -> None:
    process.last_status = status
    db.session.add(process)
    db.session.commit()


def marshall_processes(engine_settings: EngineSettingsTable, new_global_lock: bool) -> EngineSettingsTable | None:
    """Manage processes depending on the engine status.

    This function only has to act when in the transitioning fases, i.e Pausing and Starting

    Args:
        engine_settings: Engine status containing the lock and status fields
        new_global_lock: The state to which needs to be transitioned

    Returns:
        Engine status or none

    """
    try:
        # This is the first process/container that will pick up the "running" queue
        if engine_settings.global_lock and not new_global_lock:
            # Update the global lock to unlocked, to make sure no one else picks up the queue
            engine_settings.global_lock = new_global_lock
            db.session.commit()

            # Resume all the running processes
            for process in _get_running_processes():
                set_process_status(process, ProcessStatus.RESUMED)
                resume_process(process, user=SYSTEM_USER)

        elif not engine_settings.global_lock and new_global_lock:
            # Lock the engine
            logger.info("Locking the orchestrator engine, Processes will run until the next step")
            engine_settings.global_lock = new_global_lock
            db.session.commit()
        else:
            logger.info(
                "Engine is already locked or unlocked, global lock is unchanged",
                global_lock=engine_settings.global_lock,
                new_status=new_global_lock,
            )

        return engine_settings

    except SQLAlchemyError:
        logger.exception("Encountered a database error, aborting and stopping. Health check will crash the app")
        return None
    except ValueError:
        logger.exception("Encountered an anomaly, locking the engine; manual intervention necessary to fix")
        engine_settings.global_lock = True
        db.session.commit()
        return None


class ThreadPoolWorkerStatus(WorkerStatus):
    def __init__(self) -> None:
        super().__init__(executor_type="threadpool")
        thread_pool = get_thread_pool()
        self.number_of_workers_online = getattr(thread_pool, "_max_workers", -1)
        self.number_of_queued_jobs = thread_pool._work_queue.qsize() if hasattr(thread_pool, "_work_queue") else 0
        self.number_of_running_jobs = len(getattr(thread_pool, "_threads", []))



# Copyright 2019-2024 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from starlette.requests import Request
from structlog import get_logger

logger = get_logger(__name__)


def deprecated_endpoint(request: Request) -> None:
    logger.warning(
        "This function is deprecated. Please use the GraphQL query instead", method=request.method, url=str(request.url)
    )



"""Changed timestamping fields in process_steps.

Revision ID: 93fc5834c7e5
Revises: 4b58e336d1bf
Create Date: 2025-07-01 14:20:44.755694

"""

import sqlalchemy as sa
from alembic import op

from orchestrator import db

# revision identifiers, used by Alembic.
revision = "93fc5834c7e5"
down_revision = "4b58e336d1bf"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "process_steps",
        sa.Column(
            "started_at",
            db.UtcTimestamp(timezone=True),
            server_default=sa.text("statement_timestamp()"),
            nullable=False,
        ),
    )
    op.alter_column("process_steps", "executed_at", new_column_name="completed_at")
    # conn = op.get_bind()
    # sa.select
    # ### end Alembic commands ###
    # Backfill started_at field correctly using proper aliasing
    op.execute(
        """
        WITH backfill_started_at AS (
            SELECT
                ps1.stepid,
                COALESCE(prev.completed_at, p.started_at) AS new_started_at
            FROM process_steps ps1
            JOIN processes p ON ps1.pid = p.pid
            LEFT JOIN LATERAL (
                SELECT ps2.completed_at
                FROM process_steps ps2
                WHERE ps2.pid = ps1.pid AND ps2.completed_at < ps1.completed_at
                ORDER BY ps2.completed_at DESC
                LIMIT 1
            ) prev ON true
        )
        UPDATE process_steps
        SET started_at = b.new_started_at
        FROM backfill_started_at b
        WHERE process_steps.stepid = b.stepid;
    """
    )


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("process_steps", "started_at")
    op.alter_column("process_steps", "completed_at", new_column_name="executed_at")
    # ### end Alembic commands ###



from typing import Any

from orchestrator.graphql.pagination import Connection, PageInfo


def to_graphql_result_page(
    items: list[Any],
    first: int,
    after: int,
    total: int | None,
    sort_fields: list[str] | None = None,
    filter_fields: list[str] | None = None,
) -> Connection:
    has_next_page = len(items) > first

    page_items = items[:first]
    page_items_length = len(page_items)
    start_cursor = after if page_items_length else None
    end_cursor = after + page_items_length - 1

    return Connection(
        page=page_items,
        page_info=PageInfo(
            has_previous_page=bool(after),
            has_next_page=has_next_page,
            start_cursor=start_cursor,
            end_cursor=end_cursor,
            total_items=total if total else 0,
            sort_fields=sort_fields or [],
            filter_fields=filter_fields or [],
        ),
    )



import redis.asyncio
import redis.client
import redis.exceptions
from pydantic import RedisDsn
from redis import Redis
from redis.asyncio import Redis as AIORedis
from redis.asyncio.retry import Retry as AIORetry
from redis.backoff import EqualJitterBackoff
from redis.retry import Retry

from orchestrator.settings import app_settings

REDIS_RETRY_ON_ERROR = [redis.exceptions.ConnectionError]
REDIS_RETRY_ON_TIMEOUT = True
REDIS_RETRY_BACKOFF = EqualJitterBackoff(base=0.05)


def create_redis_client(redis_url: str | RedisDsn) -> redis.client.Redis:
    """Create sync Redis client for the given Redis DSN with retry handling for connection errors and timeouts."""
    return Redis.from_url(
        str(redis_url),
        retry_on_error=REDIS_RETRY_ON_ERROR,  # type: ignore[arg-type]
        retry_on_timeout=REDIS_RETRY_ON_TIMEOUT,
        retry=Retry(REDIS_RETRY_BACKOFF, app_settings.REDIS_RETRY_COUNT),
    )


def create_redis_asyncio_client(redis_url: str | RedisDsn) -> redis.asyncio.client.Redis:
    """Create async Redis client for the given Redis DSN with retry handling for connection errors and timeouts."""
    return AIORedis.from_url(
        str(redis_url),
        retry_on_error=REDIS_RETRY_ON_ERROR,  # type: ignore[arg-type]
        retry_on_timeout=REDIS_RETRY_ON_TIMEOUT,
        retry=AIORetry(REDIS_RETRY_BACKOFF, app_settings.REDIS_RETRY_COUNT),
    )



import time
from typing import Any

from sqlalchemy import Connection, event
from sqlalchemy.engine import Engine

_listener_registry = []


def monitor_sqlalchemy_queries() -> None:

    @event.listens_for(Engine, "before_cursor_execute")
    def before_cursor_execute(conn: Connection, *_args: Any) -> None:
        conn.info["queries_started"] = conn.info.get("queries_started", 0) + 1
        conn.info.setdefault("query_start_time", []).append(time.time())

    _listener_registry.append((Engine, "before_cursor_execute", before_cursor_execute))

    @event.listens_for(Engine, "after_cursor_execute")
    def after_cursor_execute(conn: Connection, *_args: Any) -> None:
        conn.info["queries_completed"] = conn.info.get("queries_completed", 0) + 1
        total = time.time() - conn.info["query_start_time"].pop(-1)
        conn.info["query_time_spent"] = conn.info.get("query_time_spent", 0.0) + total

    _listener_registry.append((Engine, "after_cursor_execute", after_cursor_execute))


def disable_listeners() -> None:
    while _listener_registry:
        listener = _listener_registry.pop()
        event.remove(*listener)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from nwastdlib.graphql.extensions.error_handler_extension import ErrorType, register_error
from orchestrator.db.filters import CallableErrorHandler
from orchestrator.graphql.types import OrchestratorInfo


def _format_context(context: dict) -> str:
    if not context:
        return ""
    return "({})".format(" ".join(f"{k}={v}" for k, v in context.items()))


def create_resolver_error_handler(info: OrchestratorInfo) -> CallableErrorHandler:
    def handle_error(message: str, **context) -> None:  # type: ignore
        return register_error(" ".join([message, _format_context(context)]), info, error_type=ErrorType.BAD_REQUEST)

    return handle_error
